from typing import Any

from ollama import Client

from not_again_ai.base.file_system import readable_size


def list_models(client: Client) -> list[dict[str, Any]]:
    """List models that are available locally.

    Args:
        client (Client): The Ollama client.

    Returns:
        list[dict[str, Any]]: A list of dictionaries (each corresponding to an available model) with the following keys:
            name (str): Name of the model
            model (str): Name of the model. This should be the same as the name.
            modified_at (str): The date and time the model was last modified.
            size (int): The size of the model in bytes.
            size_readable (str): The size of the model in a human-readable format.
            details (dict[str, Any]): Additional details about the model.
    """
    response = client.list().get("models", [])

    response_data = []
    for model_data in response:
        curr_model_data = {}
        curr_model_data["name"] = model_data["name"]
        curr_model_data["model"] = model_data["model"]
        curr_model_data["modified_at"] = model_data["modified_at"]
        curr_model_data["size"] = model_data["size"]
        curr_model_data["size_readable"] = readable_size(model_data["size"])
        curr_model_data["details"] = model_data["details"]

        response_data.append(curr_model_data)

    return response_data


def is_model_available(model_name: str, client: Client) -> bool:
    """Check if a model is available locally.

    Args:
        model_name (str): The name of the model.
        client (Client): The Ollama client.

    Returns:
        bool: True if the model is available locally, False otherwise.
    """
    # If model_name does not have a ":", append ":latest"
    if ":" not in model_name:
        model_name = f"{model_name}:latest"
    models = list_models(client)
    return any(model["name"] == model_name for model in models)


def show(model_name: str, client: Client) -> dict[str, Any]:
    """Show information about a model including the modelfile, available parameters, template, and additional details.

    Args:
        model_name (str): The name of the model.
        client (Client): The Ollama client.
    """
    response = client.show(model_name)

    response_data = {}
    response_data["modelfile"] = response["modelfile"]
    response_data["parameters"] = response["parameters"]
    response_data["template"] = response["template"]
    response_data["details"] = response["details"]
    return response_data


def pull(model_name: str, client: Client) -> Any:
    """Pull a model from the Ollama server and returns the status of the pull operation."""
    return client.pull(model_name)


def delete(model_name: str, client: Client) -> Any:
    """Delete a model from the local filesystem and returns the status of the delete operation."""
    return client.delete(model_name)



from typing import Any

from ollama import Client
from openai import OpenAI

from not_again_ai.llm.ollama import chat_completion as chat_completion_ollama
from not_again_ai.llm.openai_api import chat_completion as chat_completion_openai


def chat_completion(
    messages: list[dict[str, Any]],
    model: str,
    client: OpenAI | Client,
    max_tokens: int | None = None,
    temperature: float = 0.7,
    json_mode: bool = False,
    seed: int | None = None,
    **kwargs: Any,
) -> dict[str, Any]:
    """Creates a common wrapper around chat completion models from different providers.
    Currently supports the OpenAI API and Ollama local models.
    All input parameters are supported by all providers in similar ways and the output is standardized.

    Args:
        messages (list[dict[str, Any]]): A list of messages to send to the model.
        model (str): The model name to use.
        client (OpenAI | Client): The client object to use for chat completion.
        max_tokens (int, optional): The maximum number of tokens to generate.
        temperature (float, optional): The temperature of the model. Increasing the temperature will make the model answer more creatively.
        json_mode (bool, optional): This will structure the response as a valid JSON object.
        seed (int, optional): The seed to use for the model for reproducible outputs.

    Returns:
        dict[str, Any]: A dictionary with the following keys
            message (str | dict): The content of the generated assistant message.
                If json_mode is True, this will be a dictionary.
            completion_tokens (int): The number of tokens used by the model to generate the completion.
            extras (dict): This will contain any additional fields returned by corresponding provider.
    """
    # Determine which chat_completion function to call based on the client type
    if isinstance(client, OpenAI):
        response = chat_completion_openai.chat_completion(
            messages=messages,
            model=model,
            client=client,
            max_tokens=max_tokens,
            temperature=temperature,
            json_mode=json_mode,
            seed=seed,
            **kwargs,
        )
    elif isinstance(client, Client):
        response = chat_completion_ollama.chat_completion(
            messages=messages,
            model=model,
            client=client,
            max_tokens=max_tokens,
            temperature=temperature,
            json_mode=json_mode,
            seed=seed,
            **kwargs,
        )
    else:
        raise ValueError("Invalid client type")

    # Parse the responses to be consistent
    response_data = {}
    response_data["message"] = response.get("message", None)
    response_data["completion_tokens"] = response.get("completion_tokens", None)

    # Return any additional fields from the response in an "extras" dictionary
    extras = {k: v for k, v in response.items() if k not in response_data}
    if extras:
        response_data["extras"] = extras

    return response_data



import base64
from copy import deepcopy
import mimetypes
from pathlib import Path
from typing import Any

from liquid import Template


def _validate_message_vision(message: dict[str, list[dict[str, Path | str]] | str]) -> bool:
    """Validates that a message for a vision model is valid"""
    valid_fields = ["role", "content", "name", "tool_call_id", "tool_calls"]
    if not all(key in valid_fields for key in message):
        raise ValueError(f"Message contains invalid fields: {message.keys()}")

    valid_roles = ["system", "user", "assistant", "tool"]
    if message["role"] not in valid_roles:
        raise ValueError(f"Message contains invalid role: {message['role']}")

    if not isinstance(message["content"], list) and not isinstance(message["content"], str):
        raise ValueError(f"content must be a list of dictionaries or a string: {message['content']}")

    if isinstance(message["content"], list):
        for content_part in message["content"]:
            if isinstance(content_part, dict):
                if "image" not in content_part:
                    raise ValueError(f"Dictionary content part must contain 'image' key: {content_part}")
                if "detail" in content_part and content_part["detail"] not in ["low", "high"]:
                    raise ValueError(f"Optional 'detail' key must be 'low' or 'high': {content_part['detail']}")
            elif not isinstance(content_part, str):
                raise ValueError(f"content_part must be a dictionary or a string: {content_part}")

    return True


def encode_image(image_path: Path) -> str:
    """Encodes an image file at the given Path to base64.

    Args:
        image_path: The path to the image file to encode.

    Returns:
        The base64 encoded image as a string.
    """
    with Path.open(image_path, "rb") as image_file:
        return base64.b64encode(image_file.read()).decode("utf-8")


def create_image_url(image_path: Path) -> str:
    """Creates a data URL for an image file at the given Path.

    Args:
        image_path: The path to the image file to encode.

    Returns:
        The data URL for the image.
    """
    image_data = encode_image(image_path)

    valid_mime_types = ["image/jpeg", "image/png", "image/webp", "image/gif"]

    # Get the MIME type from the image file extension
    mime_type = mimetypes.guess_type(image_path)[0]

    # Check if the MIME type is valid
    # List of valid types is here: https://platform.openai.com/docs/guides/vision/what-type-of-files-can-i-upload
    if mime_type not in valid_mime_types:
        raise ValueError(f"Invalid MIME type for image: {mime_type}")

    return f"data:{mime_type};base64,{image_data}"


def chat_prompt(messages_unformatted: list[dict[str, Any]], variables: dict[str, str]) -> list[dict[str, Any]]:
    """Formats a list of messages for OpenAI's chat completion API,
    including special syntax for vision models, using Liquid templating.

    Args:
        messages_unformatted (list[dict[str, list[dict[str, Path | str]] | str]]):
            A list of dictionaries where each dictionary represents a message.
            `content` can be a Liquid template string or a list of dictionaries where each dictionary
            represents a content part. Each content part can be a string or a dictionary with 'image' and 'detail' keys.
            The 'image' key must be a Path or a string representing a URL. The 'detail' key is optional and must be 'low' or 'high'.
        variables: A dictionary where each key-value pair represents a variable
            name and its value for template rendering.

    Returns:
        A list which represents messages in the format that OpenAI expects for its chat completions API.
        See here for details: https://platform.openai.com/docs/api-reference/chat/create

    Examples:
        >>> # Assume cat_image and dog_image are Path objects to image files
        >>> messages = [
        ...     {"role": "system", "content": "You are a helpful assistant."},
        ...     {
        ...         "role": "user",
        ...          "content": ["Describe the animal in the image in one word.", {"image": cat_image, "detail": "low"}],
        ...     }
        ...     {"role": "assistant", "content": "{{ answer }}"},
        ...     {
        ...         "role": "user",
        ...         "content": ["What about this animal?", {"image": dog_image, "detail": "high"}],
        ...     }
        ... ]
        >>> vars = {"answer": "Cat"}
        >>> chat_prompt(messages, vars)
        [
            {"role": "system", "content": "You are a helpful assistant."},
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": "Describe the animal in the image in one word."},
                    {
                        "type": "image_url",
                        "image_url": {"url": f"data:image/jpeg;base64,<encoding>", "detail": "low"},
                    },
                ],
            },
            {"role": "assistant", "content": "Cat"},
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": "What about this animal?"},
                    {
                        "type": "image_url",
                        "image_url": {"url": f"data:image/jpeg;base64,<encoding>", "detail": "high"},
                    },
                ],
            },
        ]
    """
    messages_formatted = deepcopy(messages_unformatted)
    for message in messages_formatted:
        if not _validate_message_vision(message):
            raise ValueError()

        if isinstance(message["content"], list):
            for i in range(len(message["content"])):
                content_part = message["content"][i]
                if isinstance(content_part, dict):
                    image_path = content_part["image"]
                    if isinstance(image_path, Path):
                        temp_content_part: dict[str, Any] = {
                            "type": "image_url",
                            "image_url": {
                                "url": create_image_url(image_path),
                            },
                        }
                        if "detail" in content_part:
                            temp_content_part["image_url"]["detail"] = content_part["detail"]
                    elif isinstance(image_path, str):
                        # Assume its a valid URL
                        pass
                    else:
                        raise ValueError(f"Image path must be a Path or str: {image_path}")
                    message["content"][i] = temp_content_part
                elif isinstance(content_part, str):
                    message["content"][i] = {
                        "type": "text",
                        "text": Template(content_part).render(**variables),
                    }
        elif isinstance(message["content"], str):
            message["content"] = Template(message["content"]).render(**variables)

    return messages_formatted



from typing import Any

from openai import OpenAI


def embed_text(
    text: str | list[str],
    client: OpenAI,
    model: str = "text-embedding-3-large",
    dimensions: int | None = None,
    encoding_format: str = "float",
    **kwargs: Any,
) -> list[float] | str | list[list[float]] | list[str]:
    """Generates an embedding vector for a given text using OpenAI's API.

    Args:
        text (str | list[str]): The input text to be embedded. Each text should not exceed 8191 tokens, which is the max for V2 and V3 models
        client (OpenAI): The OpenAI client used to interact with the API.
        model (str, optional): The ID of the model to use for embedding.
            Defaults to "text-embedding-3-large".
            Choose from text-embedding-3-small, text-embedding-3-large, text-embedding-ada-002.
            See https://platform.openai.com/docs/models/embeddings for more details.
        dimensions (int | None, optional): The number of dimensions for the output embeddings.
            This is only supported in "text-embedding-3" and later models. Defaults to None.
        encoding_format (str, optional): The format for the returned embeddings. Can be either "float" or "base64".
            Defaults to "float".

    Returns:
        list[float] | str | list[list[float]] | list[str]: The embedding vector represented as a list of floats or base64 encoded string.
            If multiple text inputs are provided, a list of embedding vectors is returned.
            The length and format of the vector depend on the model, encoding_format, and dimensions.

    Raises:
        ValueError: If 'text-embedding-ada-002' model is used and dimensions are specified,
            as this model does not support specifying dimensions.

    Example:
        client = OpenAI()
        embedding = embed_text("Example text", client, model="text-embedding-ada-002")
    """
    if model == "text-embedding-ada-002" and dimensions:
        # text-embedding-ada-002 does not support dimensions
        raise ValueError("text-embedding-ada-002 does not support dimensions")

    kwargs = {
        "model": model,
        "input": text,
        "encoding_format": encoding_format,
    }
    if dimensions:
        kwargs["dimensions"] = dimensions

    response = client.embeddings.create(**kwargs)

    responses = []
    for embedding in response.data:
        responses.append(embedding.embedding)

    if len(responses) == 1:
        return responses[0]

    return responses
